/***********************************************************************************************************************************************
Macro Name:   design_biomAR.sas
Author:       Matthew Psioda
Affiliation:  Department of Biostatistics, University of North Carolina at Chapel Hill
    
**** Macro Summary:
     This macro simulates multiple studies from a binomial design and uses adaptive randomization methods.

**** Reference Paper:
      Barry, William T., et al. "The use of Bayesian hierarchical models for adaptive randomization in biomarker-driven phase II studies." 
         Journal of biopharmaceutical statistics 25.1 (2015): 66-88. 

**** Needed Macro Variables;

    (1) seed1            = <int> seed for call streaminit();
    (2) seed2            = <int> seed for call randseed();

    (3) Nmin             = <int> minimum sample size to be investigated for a study;
    (4) Nmax             = <int> maximum sample size to be investigated for a study;
    (5) numSimStudies    = <int> number of simulation studies for each investigated sample size;

                                 *** The program will loop of study sample sizes from Nmin to Nmax performing numSimStudies
                                     simulation studies at each sample size. The number of simulation studies in total will
                                     be (Nmax - Nmin + 1) * numSimStudies;


    (6) Nstart           = <int> number of subjects per treatment group/biomarker class before AR kicks in;

    (7) J0               = <int> number of treatment groups
    (8) K0               = <int> number of biomarker classes

    (9) prevBiom         = <double> K0 dimensional IML vector of biomarker class prevalences - e.g. {0.40 0.60}
   (10) pi_True          = <double> J0 x K0 dimensional IML matrix of true response probabilities - e.g. {0.25 0.25, 0.25 0.50} for J0=K0=2
   (11) pi1_Target       = <double> target acceptable response rate
   (12) pi0_Target       = <double> target unacceptable response rate

   (13) mappingMethod    = <int> 1=ratio,2=max;

   (14) priorType        = <text> prior for randomization;

                                  *** phi_rand_NI  -> non informative prior;
                                  *** phi_rand_MOD -> moderately informative prior;
                                  *** phi_rand_BAL -> balanced prior (informative);

   (15) deltaU    = <double> lower bound for posterior probability used in stopping rule for efficacy;
   (16) deltaL    = <double> upper bound for posterior probability used in stopping rule for futility;

   (17) nbi       = <int> number of burn-in samples;
   (18) nmc       = <int> number of mcmc samples;
   (19) nthin     = <int> thinning rate;


**** Needed SAS library;
     
     *** libname s "<<< storage location for simulation results >>>";

***********************************************************************************************************************************************/

%macro design_biomAR(
    seed1 = 1,
    seed2 = 2,

    Nmax   =  25,
    Nmin   =  25,       
    numSimStudies = 5,

    Nstart =   1,

    J0     =   2,
    K0     =   2,

    prevBiom   = { 0.50 0.50 },
    pi_True    = %str({ 0.25 0.25, 0.25 0.50 }),
    pi1_Target = 0.50,
    pi0_Target = 0.25,

    mappingMethod = 2,
    priorType     = phi_rand_BAL,

    deltaU =  0.900,
    deltaL =  0.025,

    nbi   =  500,
    nmc   = 5000,
    nthin =    1

);

proc IML;

call streaminit(&seed1.);
call randseed(&seed2.);
*****************************************************************************************;
** define functions needed for simulation;

     ** sample from a truncated normal (taken from Rick Wicklin blog);
     ** http://blogs.sas.com/content/iml/2013/07/24/the-truncated-normal-in-sas.html;
     ** based on inverse-CDF method;
     start randTN(N, mu, sigma, support);
        
        if support = "<0" then do;
         a = .; b = 0;
        end;
        if support = ">0" then do;
         a = 0; b = .;
        end;
        
        cdf_a = choose(a=., 0, cdf("Normal", a, mu, sigma));  
        cdf_b = choose(b=., 1, cdf("Normal", b, mu, sigma)); 
        u = j(N,1,.);
        call randgen(u, "Uniform");           
        z = cdf_a + (cdf_b - cdf_a)*u;     

        if      z > 0.99999 then x = mu + 4*sigma;
        else if z < 0.00001 then x = mu - 4*sigma;
        else x = quantile("Normal", z, mu, sigma); 
        return(x);
     finish;

     *** Gibbs sampler;
     start mcmc (N0,N1,priorParms) global(J0,K0,nbi,nmc,nthin,z_p0,z_p1,prFut,prEff,piHat,rMax);

        ** calculate total N;
        N  = N0 + N1;

        ** calculate treatment totals;
        Nj = N[,+];

        ** copy in prior parameters;
        alpha  = priorParms[1];
        sigma2 = priorParms[2];
        tau2   = priorParms[3];

        ** containers for samples;
        muSamples = J(nbi + nmc,J0*K0,0);

        ** containers for current values of parameters;
        mu       = J(J0,K0,0);
        phi      = J(J0,1,0);  
        thetaBar = J(J0,K0,0);
 
        **** Gibbs sampler;
        do i = 1 to nbi + nmc;

          ** Update sum of theta_ijk (only the sum is needed);
          do j = 1 to J0;
           do k = 1 to K0;
             thetaBar[j,k] = 0;
             if N0[j,k] > 0 then thetaBar[j,k] = (randTN(N0[j,k],mu[j,k],1,"<0"))[+];
             if N1[j,k] > 0 then thetaBar[j,k] = thetaBar[j,k] + (randTN(N1[j,k],mu[j,k],1,">0"))[+];
             thetaBar[j,k] = thetaBar[j,k]/N[j,k];
           end;
          end;

          ** Update mu_jk;
          do j = 1 to J0;
           do k = 1 to K0;
             vr      = 1/(N[j,k] + 1/sigma2);
             sd      = sqrt(vr);
             mean    = vr * ( N[j,k]*thetaBar[j,k] + phi[j]/sigma2 );

             mu[j,k] = rand('normal',mean,sd);
           end;
          end;

          ** Update phi_j;
          do j = 1 to J0;
            meanN  = tau2*(N[j,]#mu[j,])[+] + alpha;
            meanD  = Nj[j]*tau2 + 1;
            mean   = meanN/meanD;
            sd     = sqrt(1 / (Nj[j]+1/tau2));

            phi[j] = rand('normal',mean,sd);
          end;

         ** write out samples;
         muSamples[i,] = t(shape(mu,J0*K0));

        end;

        ** reduce samples to those that should be kept;
        muSamples = muSamples[do(nbi+1,nbi+nmc,nthin),];
  

        ** calculate relevant posterior quantities;
        prEff = shape((muSamples >= z_p0)[:,],J0,K0);
        prFut = shape((muSamples >= z_p1)[:,],J0,K0);
        piHat = shape((cdf('normal',muSamples,0,1))[:,],J0,K0);

        rmax = J(J0,K0,0); i=0;
        do k = 1 to K0;
         set = do(k,J0*K0,K0);
         tempSamples = muSamples[,set];
         do j = 1 to J0;
          i = i+1;
          rmax[j,k] = (muSamples[,set[j]] = tempSamples[,<>])[:];
         end;
        end;
        *print prEff,prFut,piHat,rmax;

   finish;


*****************************************************************************************;
** define parameters for simulation study;

 Nm     = &Nmax.;
 Ns     = &Nstart.;
 pi     = &pi_True.;
 J0     = &J0.;
 K0     = &K0.;

 mappingMethod = &mappingMethod.;

 piBiom = &prevBiom.;

 pi0    = &pi0_Target.; z_p0 = quantile('normal',pi0);
 pi1    = &pi1_Target.; z_p1 = quantile('normal',pi1);

 deltaL = &deltaL.;
 deltaU = &deltaU.;

 nmc    = &nmc.;
 nbi    = &nbi.;
 nthin  = &nthin.;

 phi_default = J(1,3,0); 
   phi_default[1] = (z_p0 + z_p1)/2; 
   phi_default[2] = 1.00; 
   phi_default[3] = 0.01;

 phi_rand_BAL = phi_default;

 phi_rand_MOD = phi_default; 
   phi_rand_MOD[3] = 1.00;

 phi_rand_NI  = phi_default; 
   phi_rand_NI[3]  = 100;

 phi_fut = phi_default; 
   phi_fut[1] = z_p1;

 phi_eff = phi_default; 
   phi_eff[1] = z_p0;
   phi_eff[3] = 100;

do Nm = &Nmin. to &Nmax. by 1;
do study = 1 to &numSimStudies.;

 stopArm       = J(J0,K0,0);
 stopBiomClass = J(1,K0,0);
 stopStudy     = 0;

*****************************************************************************************;
** generate initial dataset;
  N0 = J(J0,K0,0);
  N1 = J(J0,K0,0);

  do j = 1 to J0;
   do k = 1 to K0;
    y = rand('binomial',pi[j,k],Ns);
    N1[j,k] = y;
    N0[j,k] = Ns-y;
   end;
  end;
*****************************************************************************************;

  s = N0[+,+] + N1[+,+];

  prFut = .;
  prEff = .;
  piHat = .;
  rMax  = .;

  iteration = 0;
  do while(s < Nm & stopStudy = 0);
  iteration = iteration + 1;

        ** perform mcmc analysis for futility check;
        call mcmc(N0,N1,phi_fut);

        prob_fut = prFut;
        *if Nm = 5 then print prEff prFut piHat rMax;

        ** determine if 
          (1) an arm should be stopped (if not already)
          (2) a biomarker class should cease enrollment
          (3) the study should stop;

        stopArm       = stopArm + (stopArm=0)#(prFut < deltaL);
        stopBiomClass = (stopArm[+,]=J0);
        stopStudy     = (stopBiomClass[+]=K0);

        screened = 0;

        ** if the study should continue;
        if stopStudy = 0 then do;

          ** perform mcmc analysis for randomization;
          call mcmc(N0,N1,&priorType.);

          ** determine randomization probabilities;

            ** set to zero for stopped arms;
            if mappingMethod = 1 then do;
              randProb = piHat # (1-stopArm);
            end;
            else do;
              randProb = rmax # (1-stopArm);
            end;

            randProbBefore = randProb;
            ** normalize;
            do k = 1 to K0;
              if randProb[+,k]>0 then do;
                 randProb[,k] = randProb[,k] / randProb[+,k];
              end;
              else randProb[,k] = 0;
            end;
           
            k = .;
            j = .;

            ** determine which biomarker class should enroll next patient;
            biomProb = piBiom#(1-stopBiomClass);
            biomProb = biomProb / biomProb[+];

            do until(k >= 1 & k <= K0);
             call randgen(k,'Table',biomProb);
            end;

            ** determine which treatment enrolled patient receive;
            do until(j >= 1 & j <= J0);
              call randgen(j,'Table',t(randProb[,k])); 
            end;
 
          ** simulation new observation;
          y = rand('bernoulli',pi[j,k]);
          if y=0 then N0[j,k] = N0[j,k]+1;
          else N1[j,k] = N1[j,k]+1;
   
          *print j k y N0 N1 randProb[l="Before"] randProb[l='After'];
 
          ** perform mcmc analysis for efficacy check;
          call mcmc(N0,N1,phi_eff);

          s    = N0[+,+] + N1[+,+];;
          Nall = t(shape(N0 + N1,J0*K0,1));

          success = t(shape((prEff > deltaU)#(1-stopArm),J0*K0,1));
          failure = t(shape(stopArm,J0*K0,1));;

          results = Nm||study||s||Nall||success||failure||t(shape(piHat,J0*K0,1))||t(shape(randProb,J0*K0,1));

          if Nm = &Nmin. & study  = 1 & iteration = 1 then do;
           all_results = results;
          end;
          else do;
           all_results = all_results//results;
          end;

        end;

  end;
  end;

end;



   labels = {"nMax" "study" "n" 

                  %do j = 1 %to &J0; %do k = 1 %to &K0; "n&j.&k."    %end; %end;
                  %do j = 1 %to &J0; %do k = 1 %to &K0; "eff&j.&k."  %end; %end;
                  %do j = 1 %to &J0; %do k = 1 %to &K0; "fut&j.&k."  %end; %end;
                  %do j = 1 %to &J0; %do k = 1 %to &K0; "mu&j.&k."   %end; %end;
                  %do j = 1 %to &J0; %do k = 1 %to &K0; "rand&j.&k." %end; %end; };

        create s.simResults from all_results[c=labels];
          append from all_results;
        close s.simResults;


quit;



%mend design_biomAR;

